/**
 * Rational.java A class for defining rational number representations and
 * associated methods.
 */

/* This file is part of Math4J.
 * Math4J is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Math4J is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Math4J; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * 
 * Copyright 2005 Anthony Magee
 */

package org.magee.math;

// import other packages
import java.io.Serializable;

import org.magee.util.MathUtil;

/**
 * The <code>Rational</code> class provides functionality for exact rational
 * numbers. This means that a number can be represented as p/q where p and q are
 * integers. An object of type <code>Rational</code> contains two fields whose
 * type is <code>long</code>. This is not a wrapper class for a primitive.
 * <p>
 * By necessity, this class provides methods for addition, subtraction,
 * multiplication, division, exponential, and other trigonometric and
 * logarithmic functions.
 * 
 * @author Anthony W. Magee
 * @version 1.0, February 9, 2002
 */

/*
 * version changes
 */
public class Rational extends Number implements Serializable, Cloneable {

    public long numerator, denominator;

    /**
     * Constructor that defines the <code>numerator</code> and
     * <code>denominator</code> parts of the number.
     * 
     * @param numerator The numerator.
     * @param denominator The denominator.
     */
    public Rational(long numerator, long denominator) {
        if(denominator == 0L)
            throw new NumberFormatException("Cannot create a Rational object with zero as the denominator");
        else {
            this.numerator = numerator;
            this.denominator = denominator;
        }
    }

    /**
     * Computes <code>this</code>^<code>power</code>.
     * 
     * @param power The exponential power.
     * @return <code>this</code>^<code>power</code>.
     */
    public Rational pow(int power) {
        return new Rational((long) Math.pow(numerator, power), (long) Math.pow(denominator, power)).reduce();
    }

    /**
     * Computes <code>this</code>+<code>r</code>.
     * 
     * @param r The addend to be added to <code>this</code>.
     * @return <code>this</code>+<code>r</code>.
     */
    public Rational add(Rational r) {
        return new Rational(r.numerator * denominator + numerator * r.denominator, r.denominator * denominator).reduce();
    }

    /**
     * Computes <code>this</code>+<code>integer</code>.
     * 
     * @param integer The addend to be added to <code>this</code>.
     * @return <code>this</code>+<code>integer</code>.
     */
    public Rational add(long integer) {
        return add(new Rational(integer, 1L));
    }

    /**
     * Computes <code>this</code>-<code>r</code>.
     * 
     * @param r The subtrahend to be subtracted from <code>this</code>.
     * @return <code>this</code>-<code>r</code>.
     */
    public Rational subtract(Rational r) {
        return add(r.negate());
    }

    /**
     * Computes <code>this</code>-<code>integer</code>.
     * 
     * @param integer The subtrahend to be subtracted from <code>this</code>.
     * @return <code>this</code>-<code>integer</code>.
     */
    public Rational subtract(long integer) {
        return subtract(new Rational(integer, 1L));
    }

    /**
     * Computes <code>this</code>*<code>r</code>.
     * 
     * @param r The multiplicand to be multiplied with <code>this</code>.
     * @return <code>this</code>*<code>r</code>.
     */
    public Rational multiply(Rational r) {
        return new Rational(numerator * r.numerator, denominator * r.denominator).reduce();
    }

    /**
     * Computes <code>this</code>*<code>scalar</code>.
     * 
     * @param scalar The scalar to be multiplied with <code>this</code>.
     * @return <code>this</code>*<code>scalar</code>.
     */
    public Rational multiply(long scalar) {
        return multiply(new Rational(scalar, 1L));
    }

    /**
     * Computes the absolute value of <code>this</code>. sqrt(<code>this</code>^2)
     * 
     * @return sqrt(<code>this</code>^2).
     */
    public Rational abs() {
        return new Rational((numerator < 0L) ? -numerator : numerator, (denominator < 0L) ? -denominator : denominator).reduce();
    }

    /**
     * Computes <code>this</code>*<code>r</code>.
     * 
     * @param r The divisor of <code>this</code>.
     * @return <code>this</code>/<code>r</code>.
     */
    public Rational divide(Rational r) {
        return multiply(r.inverse());
    }

    /**
     * Computes <code>this</code>*<code>scalar</code>.
     * 
     * @param scalar The divisor of <code>this</code>.
     * @return <code>this</code>/<code>scalar</code>.
     */
    public Rational divide(long scalar) {
        return divide(new Rational(scalar, 1L));
    }

    /**
     * Computes the inverse of <code>this</code>.
     * 
     * @return <code>this</code>^-1.
     */
    public Rational inverse() {
        return new Rational(denominator, numerator).reduce();
    }

    /**
     * Computes the negative of this rational.
     * 
     * @return -<code>this</code>
     */
    public Rational negate() {
        return new Rational(-numerator, denominator);
    }

    /**
     * Reduces the rational to its simplest form.
     * 
     * @return This rational reduced.
     */
    public Rational reduce() {
        long[] numFactors = MathUtil.factor(numerator);
        long[] denomFactors = MathUtil.factor(denominator);

        for(int i = 0; i < numFactors.length; i++) {
            for(int j = 0; j < denomFactors.length; j++) {
                if(numFactors[i] == denomFactors[j] && numFactors[i] < 1L && denomFactors[j] != 1L) {
                    numFactors[i] = 1L;
                    denomFactors[j] = 1L;
                }
            }
        }

        long num = 1L;
        for(int i = 0; i < numFactors.length; i++) {
            num *= numFactors[i];
        }

        long denom = 1L;
        for(int i = 0; i < denomFactors.length; i++) {
            denom *= denomFactors[i];
        }

        return new Rational(num, denom);
    }

    /**
     * Returns a string representation of this rational.
     */
    public String toString() {
        return numerator + " / " + denominator;
    }

    /**
     * Returns this rational as an int.
     * 
     * @return This rational as an int.
     */
    public int intValue() {
        return ((int) numerator / (int) denominator);
    }

    /**
     * Returns this rational as an long.
     * 
     * @return This rational as an long.
     */
    public long longValue() {
        return (numerator / denominator);
    }

    /**
     * Returns this rational as an float.
     * 
     * @return This rational as an float.
     */
    public float floatValue() {
        return ((float) numerator / denominator);
    }

    /**
     * Returns this rational as an double.
     * 
     * @return This rational as an double.
     */
    public double doubleValue() {
        return ((double) numerator / denominator);
    }

    static final long serialVersionUID = 1L;
}